@@ -7,28 +7,26 @@
 //! The special case of the initialize expressions for table elements offsets or global variables
 //! is handled, according to the semantics of WebAssembly, to only specific expressions that are
 //! interpreted on the fly.
-use crate::environ::{Alias, ModuleEnvironment, WasmError, WasmResult};
+use crate::environ::{Alias, ModuleEnvironment};
 use crate::state::ModuleTranslationState;
-use crate::translation_utils::{
-    tabletype_to_type, type_to_type, DataIndex, ElemIndex, EntityIndex, EntityType, Event,
-    EventIndex, FuncIndex, Global, GlobalIndex, GlobalInit, InstanceIndex, Memory, MemoryIndex,
-    ModuleIndex, Table, TableElementType, TableIndex, TypeIndex,
-};
 use crate::wasm_unsupported;
+use crate::{
+    DataIndex, ElemIndex, EntityIndex, EntityType, FuncIndex, Global, GlobalIndex, GlobalInit,
+    InstanceIndex, Memory, MemoryIndex, ModuleIndex, Table, TableIndex, Tag, TagIndex, TypeIndex,
+    WasmError, WasmResult,
+};
 use core::convert::TryFrom;
 use core::convert::TryInto;
-use cranelift_codegen::ir::immediates::V128Imm;
-use cranelift_codegen::ir::{self, AbiParam, Signature};
 use cranelift_entity::packed_option::ReservedValue;
 use cranelift_entity::EntityRef;
 use std::boxed::Box;
 use std::vec::Vec;
 use wasmparser::{
     self, Data, DataKind, DataSectionReader, Element, ElementItem, ElementItems, ElementKind,
-    ElementSectionReader, EventSectionReader, EventType, Export, ExportSectionReader, ExternalKind,
-    FunctionSectionReader, GlobalSectionReader, GlobalType, ImportSectionEntryType,
-    ImportSectionReader, MemorySectionReader, MemoryType, NameSectionReader, Naming, Operator,
-    TableSectionReader, TableType, TypeDef, TypeSectionReader,
+    ElementSectionReader, Export, ExportSectionReader, ExternalKind, FunctionSectionReader,
+    GlobalSectionReader, GlobalType, ImportSectionEntryType, ImportSectionReader,
+    MemorySectionReader, MemoryType, NameSectionReader, Naming, Operator, TableSectionReader,
+    TableType, TagSectionReader, TagType, TypeDef, TypeSectionReader,
 };
 
 fn entity_type(
@@ -46,52 +44,38 @@ fn entity_type(
             EntityType::Instance(environ.type_to_instance_type(TypeIndex::from_u32(sig))?)
         }
         ImportSectionEntryType::Memory(ty) => EntityType::Memory(memory(ty)),
-        ImportSectionEntryType::Event(evt) => EntityType::Event(event(evt)),
-        ImportSectionEntryType::Global(ty) => {
-            EntityType::Global(global(ty, environ, GlobalInit::Import)?)
-        }
-        ImportSectionEntryType::Table(ty) => EntityType::Table(table(ty, environ)?),
+        ImportSectionEntryType::Tag(t) => EntityType::Tag(tag(t)),
+        ImportSectionEntryType::Global(ty) => EntityType::Global(global(ty, GlobalInit::Import)?),
+        ImportSectionEntryType::Table(ty) => EntityType::Table(table(ty)?),
     })
 }
 
 fn memory(ty: MemoryType) -> Memory {
-    match ty {
-        MemoryType::M32 { limits, shared } => Memory {
-            minimum: limits.initial,
-            maximum: limits.maximum,
-            shared: shared,
-        },
-        // FIXME(#2361)
-        MemoryType::M64 { .. } => unimplemented!(),
+    Memory {
+        minimum: ty.initial,
+        maximum: ty.maximum,
+        shared: ty.shared,
+        memory64: ty.memory64,
     }
 }
 
-fn event(e: EventType) -> Event {
-    Event {
+fn tag(e: TagType) -> Tag {
+    Tag {
         ty: TypeIndex::from_u32(e.type_index),
     }
 }
 
-fn table(ty: TableType, environ: &mut dyn ModuleEnvironment<'_>) -> WasmResult<Table> {
+fn table(ty: TableType) -> WasmResult<Table> {
     Ok(Table {
         wasm_ty: ty.element_type.try_into()?,
-        ty: match tabletype_to_type(ty.element_type, environ)? {
-            Some(t) => TableElementType::Val(t),
-            None => TableElementType::Func,
-        },
-        minimum: ty.limits.initial,
-        maximum: ty.limits.maximum,
+        minimum: ty.initial,
+        maximum: ty.maximum,
     })
 }
 
-fn global(
-    ty: GlobalType,
-    environ: &mut dyn ModuleEnvironment<'_>,
-    initializer: GlobalInit,
-) -> WasmResult<Global> {
+fn global(ty: GlobalType, initializer: GlobalInit) -> WasmResult<Global> {
     Ok(Global {
         wasm_ty: ty.content_type.try_into()?,
-        ty: type_to_type(ty.content_type, environ).unwrap(),
         mutability: ty.mutable,
         initializer,
     })
@@ -110,18 +94,7 @@ pub fn parse_type_section<'a>(
     for entry in types {
         match entry? {
             TypeDef::Func(wasm_func_ty) => {
-                let mut sig = Signature::new(environ.target_config().default_call_conv);
-                sig.params.extend(wasm_func_ty.params.iter().map(|ty| {
-                    let cret_arg: ir::Type = type_to_type(*ty, environ)
-                        .expect("only numeric types are supported in function signatures");
-                    AbiParam::new(cret_arg)
-                }));
-                sig.returns.extend(wasm_func_ty.returns.iter().map(|ty| {
-                    let cret_arg: ir::Type = type_to_type(*ty, environ)
-                        .expect("only numeric types are supported in function signatures");
-                    AbiParam::new(cret_arg)
-                }));
-                environ.declare_type_func(wasm_func_ty.clone().try_into()?, sig)?;
+                environ.declare_type_func(wasm_func_ty.clone().try_into()?)?;
                 module_translation_state
                     .wasm_types
                     .push((wasm_func_ty.params, wasm_func_ty.returns));
@@ -186,15 +159,15 @@ pub fn parse_import_section<'data>(
             ImportSectionEntryType::Memory(ty) => {
                 environ.declare_memory_import(memory(ty), import.module, import.field)?;
             }
-            ImportSectionEntryType::Event(e) => {
-                environ.declare_event_import(event(e), import.module, import.field)?;
+            ImportSectionEntryType::Tag(e) => {
+                environ.declare_tag_import(tag(e), import.module, import.field)?;
             }
             ImportSectionEntryType::Global(ty) => {
-                let ty = global(ty, environ, GlobalInit::Import)?;
+                let ty = global(ty, GlobalInit::Import)?;
                 environ.declare_global_import(ty, import.module, import.field)?;
             }
             ImportSectionEntryType::Table(ty) => {
-                let ty = table(ty, environ)?;
+                let ty = table(ty)?;
                 environ.declare_table_import(ty, import.module, import.field)?;
             }
         }
@@ -233,7 +206,7 @@ pub fn parse_table_section(
     environ.reserve_tables(tables.get_count())?;
 
     for entry in tables {
-        let ty = table(entry?, environ)?;
+        let ty = table(entry?)?;
         environ.declare_table(ty)?;
     }
 
@@ -255,16 +228,16 @@ pub fn parse_memory_section(
     Ok(())
 }
 
-/// Parses the Event section of the wasm module.
-pub fn parse_event_section(
-    events: EventSectionReader,
+/// Parses the Tag section of the wasm module.
+pub fn parse_tag_section(
+    tags: TagSectionReader,
     environ: &mut dyn ModuleEnvironment,
 ) -> WasmResult<()> {
-    environ.reserve_events(events.get_count())?;
+    environ.reserve_tags(tags.get_count())?;
 
-    for entry in events {
-        let event = event(entry?);
-        environ.declare_event(event)?;
+    for entry in tags {
+        let tag = tag(entry?);
+        environ.declare_tag(tag)?;
     }
 
     Ok(())
@@ -286,7 +259,7 @@ pub fn parse_global_section(
             Operator::F32Const { value } => GlobalInit::F32Const(value.bits()),
             Operator::F64Const { value } => GlobalInit::F64Const(value.bits()),
             Operator::V128Const { value } => {
-                GlobalInit::V128Const(V128Imm::from(value.bytes().to_vec().as_slice()))
+                GlobalInit::V128Const(u128::from_le_bytes(*value.bytes()))
             }
             Operator::RefNull { ty: _ } => GlobalInit::RefNullConst,
             Operator::RefFunc { function_index } => {
@@ -302,7 +275,7 @@ pub fn parse_global_section(
                 ));
             }
         };
-        let ty = global(ty, environ, initializer)?;
+        let ty = global(ty, initializer)?;
         environ.declare_global(ty)?;
     }
 
@@ -333,7 +306,7 @@ pub fn parse_export_section<'data>(
             ExternalKind::Memory => {
                 environ.declare_memory_export(MemoryIndex::new(index), field)?
             }
-            ExternalKind::Event => environ.declare_event_export(EventIndex::new(index), field)?,
+            ExternalKind::Tag => environ.declare_tag_export(TagIndex::new(index), field)?,
             ExternalKind::Global => {
                 environ.declare_global_export(GlobalIndex::new(index), field)?
             }
@@ -389,7 +362,7 @@ pub fn parse_element_section<'data>(
             } => {
                 let mut init_expr_reader = init_expr.get_binary_reader();
                 let (base, offset) = match init_expr_reader.read_operator()? {
-                    Operator::I32Const { value } => (None, value as u32 as usize),
+                    Operator::I32Const { value } => (None, value as u32),
                     Operator::GlobalGet { global_index } => {
                         (Some(GlobalIndex::from_u32(global_index)), 0)
                     }
@@ -400,12 +373,6 @@ pub fn parse_element_section<'data>(
                         ));
                     }
                 };
-                // Check for offset + len overflow
-                if offset.checked_add(segments.len()).is_none() {
-                    return Err(wasm_unsupported!(
-                        "element segment offset and length overflows"
-                    ));
-                }
                 environ.declare_table_elements(
                     TableIndex::from_u32(table_index),
                     base,
@@ -418,7 +385,7 @@ pub fn parse_element_section<'data>(
                 environ.declare_passive_element(index, segments)?;
             }
             ElementKind::Declared => {
-                // Nothing to do here.
+                environ.declare_elements(segments)?;
             }
         }
     }
@@ -441,7 +408,8 @@ pub fn parse_data_section<'data>(
             } => {
                 let mut init_expr_reader = init_expr.get_binary_reader();
                 let (base, offset) = match init_expr_reader.read_operator()? {
-                    Operator::I32Const { value } => (None, value as u32 as usize),
+                    Operator::I32Const { value } => (None, value as u64),
+                    Operator::I64Const { value } => (None, value as u64),
                     Operator::GlobalGet { global_index } => {
                         (Some(GlobalIndex::from_u32(global_index)), 0)
                     }
@@ -452,12 +420,6 @@ pub fn parse_data_section<'data>(
                         ))
                     }
                 };
-                // Check for offset + len overflow
-                if offset.checked_add(data.len()).is_none() {
-                    return Err(wasm_unsupported!(
-                        "data segment offset and length overflows"
-                    ));
-                }
                 environ.declare_data_initialization(
                     MemoryIndex::from_u32(memory_index),
                     base,
@@ -497,19 +459,31 @@ pub fn parse_name_section<'data>(
                 environ.declare_module_name(name);
             }
             wasmparser::Name::Local(l) => {
-                let mut reader = l.get_function_local_reader()?;
-                for _ in 0..reader.get_count() {
+                let mut reader = l.get_indirect_map()?;
+                for _ in 0..reader.get_indirect_count() {
                     let f = reader.read()?;
-                    if f.func_index == u32::max_value() {
+                    if f.indirect_index == u32::max_value() {
                         continue;
                     }
                     let mut map = f.get_map()?;
                     for _ in 0..map.get_count() {
                         let Naming { index, name } = map.read()?;
-                        environ.declare_local_name(FuncIndex::from_u32(f.func_index), index, name)
+                        environ.declare_local_name(
+                            FuncIndex::from_u32(f.indirect_index),
+                            index,
+                            name,
+                        )
                     }
                 }
             }
+            wasmparser::Name::Label(_)
+            | wasmparser::Name::Type(_)
+            | wasmparser::Name::Table(_)
+            | wasmparser::Name::Global(_)
+            | wasmparser::Name::Memory(_)
+            | wasmparser::Name::Element(_)
+            | wasmparser::Name::Data(_)
+            | wasmparser::Name::Unknown { .. } => {}
         }
     }
     Ok(())
@@ -539,7 +513,7 @@ pub fn parse_instance_section<'data>(
                     ExternalKind::Instance => {
                         EntityIndex::Instance(InstanceIndex::from_u32(arg.index))
                     }
-                    ExternalKind::Event => unimplemented!(),
+                    ExternalKind::Tag => unimplemented!(),
 
                     // this won't pass validation
                     ExternalKind::Type => unreachable!(),
